package de.lmu.ifi.pst.x10080133.action;

import java.awt.event.ActionEvent;
import java.lang.reflect.Array;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.JTextArea;

import com.nomagic.magicdraw.ui.browser.actions.DefaultBrowserAction;
import com.nomagic.uml2.ext.jmi.helpers.StereotypesHelper;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Element;
import com.nomagic.uml2.ext.magicdraw.mdprofiles.Stereotype;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.StateMachine;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.Transition;
import com.nomagic.uml2.ext.magicdraw.statemachines.mdbehaviorstatemachines.Vertex;
import com.nomagic.uml2.impl.magicdraw.classes.mdkernel.InstanceSpecificationImpl;
import com.nomagic.uml2.impl.magicdraw.statemachines.mdbehaviorstatemachines.FinalStateImpl;
import com.nomagic.uml2.impl.magicdraw.statemachines.mdbehaviorstatemachines.RegionImpl;
import com.nomagic.uml2.impl.magicdraw.statemachines.mdbehaviorstatemachines.StateImpl;
import com.nomagic.uml2.impl.magicdraw.statemachines.mdbehaviorstatemachines.StateMachineImpl;

import de.lmu.ifi.pst.x10080133.domain.NavigationFile;
import de.lmu.ifi.pst.x10080133.domain.NavigationNode;
import de.lmu.ifi.pst.x10080133.domain.Rule;

/**
 * Behavior of the GenerateMonitorFile context
 * 
 * @author Roman.Schwienbacher
 */
public class GenerateMonitorFileAction extends DefaultBrowserAction {

	private static final long serialVersionUID = 8830453962543810758L;

	public final static String SNP_PLUGIN_ACTION_NAME = "Generate-Monitor-File";
	public final static String SNP_PLUGIN_ACTION_ID = "GMF";

	private boolean homeProcessed = false;
	private boolean homeProcessedMoreThanOne = false;

	/**
	 * sets action-name and id by default constructor
	 */
	public GenerateMonitorFileAction() {
		super(SNP_PLUGIN_ACTION_ID, SNP_PLUGIN_ACTION_NAME, null, null);
	}

	/**
	 * scans the whole state machine and exports the navigational rule file including SNPs
	 */
	@SuppressWarnings("unchecked")
	@Override
	public void actionPerformed(ActionEvent e) {

		String stateMachineName = "";
		String applicationName = "";
		String defaultViolation = "";
		String result = "";

		homeProcessed = false;
		homeProcessedMoreThanOne = false;

		try {

			List<StateImpl> states = new ArrayList<StateImpl>();
			List<NavigationNode> navigationNodes = new ArrayList<NavigationNode>();

			// retrieve selected state-machine with corresponding states and transitions
			StateMachineImpl stateMachine = (StateMachineImpl) this.getSelectedObject();
			stateMachineName = stateMachine.getName();

			// initialize root state and first level-states
			for (Element element_lvl1 : stateMachine.getOwnedElement()) {

				if (element_lvl1 instanceof RegionImpl) { // get region of selected state machine

					for (Element element_lvl2 : element_lvl1.getOwnedElement()) {

						if (element_lvl2 instanceof StateImpl && !(element_lvl2 instanceof FinalStateImpl) && !((StateImpl) element_lvl2).getName().isEmpty()) { // get root-state

							StateImpl rootState = (StateImpl) element_lvl2;

							Collection<Stereotype> stereotypes = StereotypesHelper.getStereotypes((Element) rootState); // read tags of root-state for default-violation
							for (Stereotype stereotype : stereotypes) {
								List<Object> tagValues1 = StereotypesHelper.getStereotypePropertyValue(rootState, stereotype, "unauthorizedAccess");
								for (Object tagValue : tagValues1) {
									if (tagValue instanceof StateImpl) {
										defaultViolation = ((StateImpl) tagValue).getName();
										break;
									}
								}
							}

							applicationName = rootState.getName(); // write application name

							for (Element element_lvl3 : rootState.getOwnedElement()) {

								if (element_lvl3 instanceof RegionImpl) { // get region of root-state

									for (Element element_lvl4 : element_lvl3.getOwnedElement()) { // iterate over elements of root-state (get first level states here)

										if (element_lvl4 instanceof StateImpl) {
											states.add((StateImpl) element_lvl4);
										}
									}
								}
							}
						}
					}
				}
			}

			navigationNodes = buildNavigationNodes(states, defaultViolation, -1, false, new ArrayList<String>(), new ArrayList<String>(), ""); // create navigation nodes

			// create final navigationFile if constraints are o.k.
			if (!navigationNodes.isEmpty() && !applicationName.isEmpty() && !defaultViolation.isEmpty() && homeProcessed && !homeProcessedMoreThanOne) {

				NavigationFile navigationFile = new NavigationFile();
				navigationFile.set_comment("Build time: " + new SimpleDateFormat("dd.MM.yyyy hh:mm:ss").format(new Date()));
				navigationFile.setApplication(applicationName);
				navigationFile.setLocations(navigationNodes.toArray((NavigationNode[]) Array.newInstance(NavigationNode.class, 0))); // add navigationNodes to navigationFile
				navigationFile.setDefault_violation(defaultViolation);

				result = navigationFile.encodeToJSON();

			} else { // otherwise throw corresponding exception
				if (navigationNodes.isEmpty()) {
					throw new Exception("no navigation-nodes found");
				} else if (applicationName.isEmpty()) {
					throw new Exception("no application-name given");
				} else if (defaultViolation.isEmpty()) {
					throw new Exception("no default-violation defined");
				} else if (!homeProcessed) {
					throw new Exception("no home-state found");
				} else if (homeProcessedMoreThanOne) {
					throw new Exception("only one home-state tag per state-machine allowed");
				}
			}

		} catch (Exception _e) {
			result = "Error at generating monitor file: " + _e.getMessage();
		}

		// build result for java.properties
		result = "#PROPERTIES for AppMonitor\nnavigation.file=" + result.replaceAll(",\"application", ", \\\\\n \"application").replaceAll(",\"locations", ", \\\\\n \"locations").replaceAll("\\{\"location", " \\\\\n\t\\{\"location").replaceAll(",\"default_violation", ", \\\\\n \"default_violation");

		// build gui-response window
		JTextArea textArea = new JTextArea(result);
		textArea.setColumns(110);
		textArea.setLineWrap(true);
		textArea.setWrapStyleWord(true);
		textArea.setSize(textArea.getPreferredSize().width, 1);
		JOptionPane.showMessageDialog(null, textArea, "Generated Monitor-File for State-Machine " + stateMachineName, JOptionPane.PLAIN_MESSAGE);
	}

	/**
	 * 
	 * returns list of navigation-node instances by given states, called recursively using params as accumulator
	 * 
	 * @param states
	 * @param defaultViolation
	 * @param startStateIndex
	 * @param isHome
	 * @param roles
	 * @param incomingStateNames
	 * @param violation
	 * @return
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	private List<NavigationNode> buildNavigationNodes(List<StateImpl> states, String defaultViolation, Integer startStateIndex, boolean isHome, List<String> roles, List<String> incomingStateNames, String violation) throws Exception {

		List<NavigationNode> navigationNodes = new ArrayList<NavigationNode>();

		int localStateIndex = 0;
		
		for (StateImpl state : states) { // iterate over first level states and create List of NavigationNodes

			boolean isHomeLocal = false;
			List<String> rolesLocal = new ArrayList<String>();
			String violationLocal = "";
			Integer startStateIndexLocal = -1;

			// read incoming state names for actual state
			List<String> incomingStateNamesLocal = readIncomingStateNames(state);

			// retrieve tags for actual state
			Collection<Stereotype> stereotypes = StereotypesHelper.getStereotypes((Element) state);
			for (Stereotype stereotype : stereotypes) {
				List<Object> tagValues1 = StereotypesHelper.getStereotypePropertyValue(state, stereotype, "isHome");
				for (Object tagValue : tagValues1) {
					isHomeLocal = Boolean.parseBoolean(tagValue.toString());
					break;
				}

				List<Object> tagValues2 = StereotypesHelper.getStereotypePropertyValue(state, stereotype, "roles");
				for (Object tagValue : tagValues2) {
					if (tagValue instanceof InstanceSpecificationImpl) {
						rolesLocal.add(((InstanceSpecificationImpl) tagValue).getName());
					}
				}

				List<Object> tagValues3 = StereotypesHelper.getStereotypePropertyValue(state, stereotype, "unauthorizedAccess");
				for (Object tagValue : tagValues3) {
					if (tagValue instanceof StateImpl) {
						violationLocal = ((StateImpl) tagValue).getName();
						break;
					}
				}
			}

			if (isHomeLocal && !homeProcessed) { // check if home tag
				homeProcessed = true;
			} else if (isHomeLocal && homeProcessed) { // multiple home-state tags not allowed, so break
				homeProcessedMoreThanOne = true;
				break;
			}

			// generate navigationNodes (location level)

			// check if state contains sub-states 
			List<StateImpl> subStates = new ArrayList<StateImpl>();
			int i = 0;
			for (Element element_lvl1 : state.getOwnedElement()) {

				if (element_lvl1 instanceof RegionImpl) { // if state owns a region, it mostly has substates

					for (Element element_lvl2 : element_lvl1.getOwnedElement()) { // now get all substates by region

						if (element_lvl2 instanceof StateImpl) {

							StateImpl secondLevelState = (StateImpl) element_lvl2;

							subStates.add(secondLevelState);

							for (Transition transition : secondLevelState.getIncoming()) {

								if (transition.getSource().getHumanType().toLowerCase().startsWith("pseudo")) { // secondLevelState is start-state of sup-state, so save position
									startStateIndexLocal = i;
									break;
								}
							}

							++i;
						}
					}
				}
			}

			if (subStates.isEmpty()) { // state has no substates, so create NavigationNode instance, BREAK CONDITION

				boolean isStartState = localStateIndex == startStateIndex; // check if navigation state is lowest start state
								
				NavigationNode nn = new NavigationNode();
				nn.setHome(isStartState ? (isHome || isHomeLocal) : isHomeLocal);
				nn.setLocation(state.getName());
				nn.setViolation((violationLocal.isEmpty() && violation.isEmpty()) ? defaultViolation : (violationLocal.isEmpty() ? violation : violationLocal)); // get violation by: 1st priority local violation, 2nd priority inherited violation, 3rd priority default violation

				if (isStartState) { // if state is start-state of one or many sup-states, inherit incomingStateNames of sup-state and add all to incomingStateNamesLocal
					incomingStateNamesLocal.addAll(incomingStateNames);
				}
				
				rolesLocal = rolesLocal.isEmpty() ? roles : rolesLocal; // inherit roles if no local roles defined
				
				List<Rule> rules = new ArrayList<Rule>(); // generate rules by roles and secondLevelIncomingStateNames
				if (!nn.isHome()) { // don't write rules if it's the home state (the login view must be accessible by any user with any role)
					for (String role : rolesLocal) {
						rules.add(new Rule(role, incomingStateNamesLocal.toArray((String[]) Array.newInstance(String.class, 0))));
					}
				}

				if (nn.isHome()) { // if navigation node is hom add role-and-preVisited-free-access
					rules.add(new Rule("*", new String[] {}));
				} else if (rules.isEmpty()) { // add role-free-access if no rule given
					rules.add(new Rule("*", incomingStateNamesLocal.toArray((String[]) Array.newInstance(String.class, 0))));
				}

				nn.setRules(rules.toArray((Rule[]) Array.newInstance(Rule.class, 0)));
				navigationNodes.add(nn); // add instance of NavigationNode instance 'nn' to navigationNodes
			} 
			else { // state has substates

				// the following two values are inherited and can be accumulated (for start state) 
				
				// accumulate incoming state names with incomungStateNamesLocal for start state
				List<String> subIncomingStateNames = new ArrayList<String>();
				subIncomingStateNames.addAll(incomingStateNamesLocal);
				subIncomingStateNames.addAll(incomingStateNames);
				
				// inherit isHome for start state				
				Boolean subIsHome = isHome || isHomeLocal; 					
				
				
				// the following two values are inherited, but can be overwritten
				
				// if violationLocal given, overwrite inherited violation
				String subViolation = violationLocal.isEmpty() ? violation : violationLocal;
				
				// overwrite roles for start state by rolesLocal if not empty
				List<String> subRoles = new ArrayList<String>();
				subRoles.addAll(rolesLocal.isEmpty() ? roles : rolesLocal); 

				// RECURSIVE call,
				// build navigation nodes by subStates and accumulate, inherit defaultViolation, inherit isHome for start state, inherit or overwrite roles for start state
				navigationNodes.addAll(buildNavigationNodes(subStates, defaultViolation, startStateIndexLocal, subIsHome, subRoles, subIncomingStateNames, subViolation));
			}
			
			++localStateIndex;			
		}

		return navigationNodes;
	}

	/**
	 * reads all possible incoming state names to lowest level (recursive)
	 * 
	 * @param state
	 * @return
	 */
	private List<String> readIncomingStateNames(StateImpl state) {

		List<String> incomingStateNames = new ArrayList<String>(); // read incoming state names

		for (Transition transition : state.getIncoming()) { // iterate over all incoming transitions and add lowest level incoming state names
			incomingStateNames.addAll(readLowestLevelChilds(transition.getSource()));
		}

		return incomingStateNames;
	}

	/**
	 * reads all childs of given transitionSource on lowest level recursively
	 * 
	 * @param transitionSource
	 * @return
	 */
	private List<String> readLowestLevelChilds(Vertex transitionSource) {

		List<String> lowestLevelChilds = new ArrayList<String>();

		// check if incoming state contains sub-states
		boolean hasSubStates = false;

		for (Element element_lvl1 : transitionSource.getOwnedElement()) {

			if (element_lvl1 instanceof RegionImpl) { // if incoming state owns a region, it mostly has substates

				for (Element element_lvl2 : element_lvl1.getOwnedElement()) {

					if (element_lvl2 instanceof StateImpl) { // in case of incoming state has substates, set all substates as incoming states for actual state

						hasSubStates = true;

						StateImpl secondLevelState = (StateImpl) element_lvl2;

						lowestLevelChilds.addAll(readLowestLevelChilds(secondLevelState)); // incoming state has substates, so read them: RECURSIVE CALL
					}
				}
			}
		}

		// if state hasn't substates, return itself (if not empty), BREAK CONDITION for recursive method call
		if (!hasSubStates && !transitionSource.getName().isEmpty()) {
			lowestLevelChilds.add(transitionSource.getName());
		}

		return lowestLevelChilds;
	}

	@Override
	/**
	 * enable this action if selectedObject is a state-machine
	 */
	public void updateState() {
		if (this.getSelectedObject() != null) {

			if (this.getSelectedObject() instanceof StateMachine) {
				setEnabled(true);
			} else {
				setEnabled(false);
			}
		} else {
			setEnabled(false);
		}
	}
}
